//////////////////////////////////////////////////////////////////////////////
// GpgaDefaultOperators.cpp- operator classes for default implementation
//                            (32-bit binary chromosome representation with
//                            real-number (double) fitness
//////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "math.h"
#include "GpgaIndividual.h"
#include "GpgaPopulation.h"
#include "GpgaDefaultOperators.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


// define PI constant
#define PI	3.141592653589793



//////////////////////////////////////////////////////////////////////////////
// Initialization operator class

//////////////////////////////////////////////////////////////////////////////
// constructor

CDefaultInitOperator::CDefaultInitOperator(CPopulation<DWORD, double>& Population, COperatorBase& EvalOperator)
:	m_Population(Population), 
	m_EvalOperator(EvalOperator)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// destructor

CDefaultInitOperator::~CDefaultInitOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CDefaultInitOperator::Run()
{
	// init random number generator
	srand(::GetTickCount());

	CreateRandomIndividuals();
	m_EvalOperator.Run();

	return 1;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// CreateRandomIndividuals	- creates random chromosomes for all individuals
//							  in the population

void CDefaultInitOperator::CreateRandomIndividuals()
{
	// get population size
	DWORD dwPopulationSize = m_Population.GetPopulationSize();
	// get individual chromosome dimension
	DWORD dwChromDim = (m_Population.GetIndividualPtr(0))->GetChromosomeDim();

	for (DWORD i = 0; i < dwPopulationSize; ++i) {
		CIndividual<DWORD, double>* pInd = m_Population.GetIndividualPtr(i);

		for (DWORD j = 0; j < dwChromDim; ++j) {
			(pInd->GetChromosomePtr())[j] = 
				((rand() & 0x00FF) |
				((rand() & 0x00FF) << 8) |
				((rand() & 0x00FF) << 16) |
				((rand() & 0x00FF) << 24));
		}
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Slave operator class

//////////////////////////////////////////////////////////////////////////////

CDefaultSlaveOperator::CDefaultSlaveOperator(COperatorBase& Selection, DWORD dwSlaveIterations)
:	m_Selection(Selection),
	m_dwSlaveIterations(dwSlaveIterations)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////

CDefaultSlaveOperator::~CDefaultSlaveOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator, runs slave operator (for 
//		  single thread config)
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CDefaultSlaveOperator::Run()
{
	return Run(0, 1);
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator, runs slave operator (for 
//		  multi-thread config)
//
// DWORD dwThreadNo		- thread number (zero-based)
// DWORD dwTotalThreads	- total number of threads
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CDefaultSlaveOperator::Run(DWORD dwThreadNo, DWORD dwTotalThreads)
{
	for (DWORD i = 0; i < m_dwSlaveIterations; ++i) {
		m_Selection.Run();
	}

	return 1;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Evaluation operator class
//
// This operator class contains several test functions (see XXXX function)

//////////////////////////////////////////////////////////////////////////////
// constructor

CDefaultEvalOperator::CDefaultEvalOperator(CPopulation<DWORD, double>& Population, double* pdLowerLimits, double* pdUpperLimits, int nEvalFnID, int nFitnessFactor)
:	m_Population(Population), 
	m_pdLowerLimits(pdLowerLimits), 
	m_pdUpperLimits(pdUpperLimits), 
	m_nEvalFnID(nEvalFnID), 
	m_nFitnessFactor(nFitnessFactor)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// destructor

CDefaultEvalOperator::~CDefaultEvalOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator, evaluates the entire population
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CDefaultEvalOperator::Run()
{
	DWORD	dwRet = 1;
	for (DWORD i = 0; i < m_Population.GetPopulationSize(); ++i) {
		dwRet &= Run(reinterpret_cast<CIndividual<class ChromClass, class FitClass>*>(m_Population.GetIndividualPtr(i)));
	}

	return dwRet;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator, evaluates the given individual
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CDefaultEvalOperator::Run(CIndividual<class ChromClass, class FitClass>* pIndividual)
{
	CIndividual<DWORD, double>* pInd = reinterpret_cast<CIndividual<DWORD, double>*>(pIndividual);
	double* pdPoint = NULL;

	try {
		pdPoint = Decode(pInd);
		pInd->SetFitness(Evaluate(pdPoint) * m_nFitnessFactor);

		delete pdPoint;

		return 1;
	} catch (...) {

		if (pdPoint) delete pdPoint;

		return 0xFFFFFFFF;
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Decode	- decodes the individual
//
// returns double*	- decoded point vector

double* CDefaultEvalOperator::Decode(CIndividual<DWORD, double>* pIndividual)
{
	DWORD dwDim = pIndividual->GetChromosomeDim();
	double* pdPoint = new double[dwDim];
	double dDelta;

	for (DWORD dwCount = 0; dwCount < dwDim; dwCount++) {
		dDelta = m_pdUpperLimits[dwCount] - m_pdLowerLimits[dwCount];
		pdPoint[dwCount] = dDelta / 0xFFFFFFFF * (pIndividual->GetChromosomePtr())[dwCount] + m_pdLowerLimits[dwCount];
	}

	return pdPoint;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Evaluate	- evaluates given vector using fn ID
//
// double* pdPoint	- point vector
//
// returns double	- function value at the given point

double CDefaultEvalOperator::Evaluate(double* pdPoint)
{
	// function dimension
	int		nDim;
	// function No.
	int		nFunctionNo;

	// get fn ID and dimesion
	nDim		= m_nEvalFnID / 1000;
	nFunctionNo	= abs(m_nEvalFnID) % 1000;

	// 1-D functions
	if (nDim == 1) {
		double x = pdPoint[0];

		switch(nFunctionNo) { 
			case 0	: return(x*sin(10*PI*x)+1.0); 

			case 1	: return(sin(10*PI*x*x)/x);
			
			case 2	: return(x*x*sin(10*PI*x*x));
			
			case 3	: return(pow(x,sin(x)));
			
			case 4	: return(x*cos(tan(x)));
			
			case 5	: return(x*cos(tan(pow(x,0.5))));
			
			case 6	: return(x*sin(30*(PI/2-atan(x))/x));
			
			case 7	: return(0.5-(sin(x)*sin(x)-0.5)/(1+0.001*x*x)/(1+0.001*x*x));
		
			default	: throw OperatorException();
		}
	// 2-D functions
	} else if (nDim == 2) {
		double x = pdPoint[0];
		double y = pdPoint[1];

		switch(nFunctionNo) {
			// Vanderplaats's 2-D function
			case 0	: return(10*pow(x,4)-20*x*x*y+10*y*y+x*x-2*x+5);

			// "egg-crate" function
			case 1	: return(x*x+y*y+25*(sin(x)*sin(x)+sin(y)*sin(y)));

			// DeJongov's test function No. 2
			case 2	: return(100*(x*x-y)*(x*x-y)+(1-x)*(1-x));

			// Schafer's test functions
			case 3	: return(0.5+pow(sin((sqrt(x*x+y*y)-0.5)/pow(1.0+0.001*(x*x+y*y), 2)), 2));

			case 4	: return(pow(x*x+y*y, 0.25)*(pow(sin(50*pow(x*x+y*y, 0.1)), 2)+1.0));

			default	: throw OperatorException();
		}
	// N-D functions
	} else {
		int		nCount;
		double	dSum;

		nDim = abs(nDim);

		switch(nFunctionNo) {
			// DeJong's test function No. 1 (usually 3-D)
			case 0	: {
				dSum = 0;
				for (nCount=0; nCount<nDim; nCount++) {
					dSum += pdPoint[nCount]*pdPoint[nCount];
				}

				return dSum;
			}

			// DeJong's test function No. 3 (usually 5-D)
			case 1	: {
				dSum = 25;
				for (nCount=0; nCount<nDim; nCount++) {
					dSum += floor(pdPoint[nCount]);
				}

				return dSum;
			}

			// Rastring's test function
			case 2	: {
				dSum = nDim*10;
				for (nCount=0; nCount<nDim; nCount++) {
					dSum += pdPoint[nCount]*pdPoint[nCount]-cos(2*PI*pdPoint[nCount]);
				}

				return dSum;
			}

			// Schwefel's test function
			case 3	: {
				dSum = 0;
				for (nCount=0; nCount<nDim; nCount++) {
					dSum += -pdPoint[nCount]*sin(sqrt(fabs(pdPoint[nCount])));
				}

				return dSum;
			}

			// Griewank's test function
			case 4	: {
				double	dProd = 1;

				dSum = 1;
				for (nCount=0; nCount<nDim; nCount++) {
					dSum += pdPoint[nCount]*pdPoint[nCount]/4000;
				}

				for (nCount=0; nCount<nDim; nCount++) {
					dProd *= cos(pdPoint[nCount]/sqrt(nCount+1));
				}

				dSum -= dProd;

				return dSum;
			}

			// Rosenbrock's test function
			case 5	: {
				dSum = 0;
				for (nCount=0; nCount<nDim-1; nCount++) {
					dSum += 100*pow((pdPoint[nCount+1]-pdPoint[nCount]*pdPoint[nCount]), 2)+pow((1-pdPoint[nCount]), 2);
				}

				return dSum;
			}

			default	: throw OperatorException();
		}
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Selection operator class
//
// Performs a 3-way tournament selection. Intended for use in slave threads, 
// therefore, it's thread safe (locks individuals for use)

//////////////////////////////////////////////////////////////////////////////
// Constructor

CDefaultSelectionOperator::CDefaultSelectionOperator(CPopulation<DWORD, double>& Population, CEvalOperatorBase& EvalOperator)
:	m_Population(Population), 
	m_EvalOperator(EvalOperator)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Destructor

CDefaultSelectionOperator::~CDefaultSelectionOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CDefaultSelectionOperator::Run()
{
	CIndividual<DWORD, double>* pIndividual1;
	CIndividual<DWORD, double>* pIndividual2;
	CIndividual<DWORD, double>* pIndividual3;
	CIndividual<DWORD, double>* pTmpInd;

	DWORD	dwChromosome1;
	DWORD	dwChromosome2;
	DWORD	dwChromosome3;

	try {
		// select first random chromosome index
		do {
			dwChromosome1 = (DWORD) (rand() % m_Population.GetPopulationSize());
		} while(!m_Population.TryLockIndividual(dwChromosome1));

		// select second random chromosome index 
		// must be different from the first one - we ensure this by 
		// thread-safe access to individuals
		do {
			dwChromosome2 = (DWORD) (rand() % m_Population.GetPopulationSize());
		} while(!m_Population.TryLockIndividual(dwChromosome2));

		// select third random chromosome index
		// must be different from the previous two - we ensure this by 
		// thread-safe access to individuals
		do {
			dwChromosome3 = (DWORD) (rand() % m_Population.GetPopulationSize());
		} while(!m_Population.TryLockIndividual(dwChromosome3));

		// now, sort the chromosomes
		pIndividual1 = m_Population.GetIndividualPtr(dwChromosome1);
		pIndividual2 = m_Population.GetIndividualPtr(dwChromosome2);
		pIndividual3 = m_Population.GetIndividualPtr(dwChromosome3);

		if (pIndividual1->GetFitness() < pIndividual2->GetFitness()) {
			pTmpInd = pIndividual1;
			pIndividual1 = pIndividual2;
			pIndividual2 = pTmpInd;
		}

		if (pIndividual2->GetFitness() < pIndividual3->GetFitness()) {
			pTmpInd = pIndividual2;
			pIndividual2 = pIndividual3;
			pIndividual3 = pTmpInd;
		}

		if (pIndividual1->GetFitness() < pIndividual2->GetFitness()) {
			pTmpInd = pIndividual1;
			pIndividual1 = pIndividual2;
			pIndividual2 = pTmpInd;
		}

		// perform the crossover
		CDefaultCrossoverOperator opCrossover(pIndividual1, pIndividual2, pIndividual3);
		opCrossover.Run();

		// now, if the offspring is the same as one of the parents, create
		// a new, random offspring chromosome
		DWORD dwChromosomeDim = pIndividual3->GetChromosomeDim();

		if ((!memcmp(pIndividual3->GetChromosomePtr(), 
					pIndividual1->GetChromosomePtr(), 
					dwChromosomeDim * sizeof(DWORD))) ||
			(!memcmp(pIndividual3->GetChromosomePtr(), 
					pIndividual2->GetChromosomePtr(), 
					dwChromosomeDim * sizeof(DWORD)))) {

			for (DWORD i = 0; i < dwChromosomeDim; ++i) {
				(pIndividual3->GetChromosomePtr())[i] = 
					((rand() & 0x00FF) |
					((rand() & 0x00FF) << 8) |
					((rand() & 0x00FF) << 16) |
					((rand() & 0x00FF) << 24));
			}
		}

		// evaluate the child
		m_EvalOperator.Run(reinterpret_cast<CIndividual<class ChromClass, class FitClass>*>(pIndividual3));
		// unlock the individuals
		m_Population.UnlockIndividual(dwChromosome1);
		m_Population.UnlockIndividual(dwChromosome2);
		m_Population.UnlockIndividual(dwChromosome3);

		return 1;
	} catch (...) {

		// unlock the individuals
		m_Population.UnlockIndividual(dwChromosome1);
		m_Population.UnlockIndividual(dwChromosome2);
		m_Population.UnlockIndividual(dwChromosome3);
		return 0xFFFFFFFF;
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Crossover operator class
//
// Performs a uniform crossover

//////////////////////////////////////////////////////////////////////////////
// Constructor

CDefaultCrossoverOperator::CDefaultCrossoverOperator(CIndividual<DWORD, double>* pParent1, CIndividual<DWORD, double>* pParent2, CIndividual<DWORD, double>* pOffspring)
:	m_pParent1(pParent1), 
	m_pParent2(pParent2), 
	m_pOffspring(pOffspring)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Destructor

CDefaultCrossoverOperator::~CDefaultCrossoverOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CDefaultCrossoverOperator::Run()
{
	try {
		DWORD	dwMask;

		DWORD	dwDim		= m_pParent1->GetChromosomeDim();
		DWORD*	pdwParent1	= m_pParent1->GetChromosomePtr();
		DWORD*	pdwParent2	= m_pParent2->GetChromosomePtr();
		DWORD*	pdwOffspring= m_pOffspring->GetChromosomePtr();

		for (DWORD i = 0; i < dwDim; ++i) {
			// for each DWORD in chromosome array, create a random mask...
			dwMask = ((rand() & 0x00FF) |
					 ((rand() & 0x00FF) << 8) |
					 ((rand() & 0x00FF) << 16) |
					 ((rand() & 0x00FF) << 24));

			// ... and do a uniform crossover
			// offspring = (parent1 & mask) | (parent2 & ~mask)
			pdwOffspring[i] =	(pdwParent1[i] & dwMask) |
								(pdwParent2[i] & ~dwMask);
		}

		return 1;
	} catch (...) {
		return 0xFFFFFFFF;
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Mutation operator class
//
// Performs a simple mutation on a population

//////////////////////////////////////////////////////////////////////////////
// Constructor

CDefaultMutationOperator::CDefaultMutationOperator(CPopulation<DWORD, double>& Population, CEvalOperatorBase& EvalOperator, double dBitMutationRate)
:	m_Population(Population), 
	m_EvalOperator(EvalOperator), 
	m_dBitMutationRate(dBitMutationRate)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Destructor

CDefaultMutationOperator::~CDefaultMutationOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CDefaultMutationOperator::Run()
{
	try {
		// determine the number of mutating chromosomes
		// we need this +0.5 in order to ensure correct truncation 
		// of the float value
		DWORD dwMutatingChromosomes;
		dwMutatingChromosomes = (DWORD)(m_Population.GetPopulationSize() * 
								m_Population.GetIndividualPtr(0)->GetChromosomeDim() *
								m_dBitMutationRate + 0.5);

		DWORD	dwBestIndividual = m_Population.GetBestIndividualIndex();
 
		// now, select dwMutatingChromosomes chromosomes to mutate, and invert one bit
		// chosen at random
		DWORD	dwRandChromosome;
		DWORD	dwRandDword;
		DWORD	dwRandomBit;
		DWORD	dwMutationMask;
		CIndividual<DWORD, double>* pMutatingInd;

		for (DWORD i = 0; i < dwMutatingChromosomes; ++i) {
			// select random chromosome 
			// (we must not select the best individual)
			do {
				dwRandChromosome = rand() % m_Population.GetPopulationSize();
			} while (dwRandChromosome == dwBestIndividual);

			// select random bit in the chromosome
			dwRandomBit = rand() % (m_Population.GetIndividualPtr(0)->GetChromosomeDim() * 32);

			// now, determine in which DWORD within the chormosome the 
			// mutating bit lies (divide by 32)...
			dwRandDword = dwRandomBit >> 5;
			// ... create the mutation mask
			dwMutationMask = 1 << (dwRandomBit % 32);
			// ... mutate the chromosome...
			pMutatingInd = m_Population.GetIndividualPtr(dwRandChromosome);
			(pMutatingInd->GetChromosomePtr())[dwRandDword] ^= dwMutationMask;

			// ... and evaluate the new individual
			m_EvalOperator.Run(reinterpret_cast<CIndividual<class ChromClass, class FitClass>*>(pMutatingInd));
		}

		// done! return.
		return 1;

	} catch (...) {
		return 0xFFFFFFFF;
	}
}

//////////////////////////////////////////////////////////////////////////////
