//////////////////////////////////////////////////////////////////////////////
// GpgaEngine.cpp	- GPGA engine class
//////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "GpgaIndividual.h"
#include "GpgaPopulation.h"
#include "GpgaOperatorBase.h"
#include "GpgaEngine.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//////////////////////////////////////////////////////////////////////////////
// CGpgaEngine	- GPGA engine class
// The class incorporates GPGA functionality.

//////////////////////////////////////////////////////////////////////////////
// Constructor	- inits the operators, and runs the init operator

CGpgaEngine::CGpgaEngine(COperatorBase& opInitialize, COperatorBase& opMaster, CSlaveOperatorBase& opSlave, DWORD dwSlaveThreads)
:	m_opInitialize(opInitialize), 
	m_opMaster(opMaster), 
	m_opSlave(opSlave), 
	m_dwSlaveThreads(dwSlaveThreads)
{
	DWORD i;

	// first, get the frequency of the high-resolution performance counter
	::QueryPerformanceFrequency(&m_llCounterFreq);
	// scale down to milliseconds
	m_llCounterFreq.QuadPart /= 1000;

	// time the initialization...
	::QueryPerformanceCounter(&m_llStartTime);

	// run the init operator
	m_opInitialize.Run();

	// create thread start events
	m_phThreadsStart	= new HANDLE[m_dwSlaveThreads];
	// create exit semaphore
	m_hThreadsExitSemaphore	= ::CreateSemaphore(NULL, 0, m_dwSlaveThreads, NULL);

	// create thread structs
	m_pThreadsData = new THREAD_DATA[m_dwSlaveThreads];
	::ZeroMemory(m_pThreadsData, m_dwSlaveThreads * sizeof(THREAD_DATA));

	// this array is used to wait for slaves to finish running slave operator
	m_phThreadsFinished	= new HANDLE[m_dwSlaveThreads];

	for (i = 0; i < m_dwSlaveThreads; ++i) {
		m_pThreadsData[i].pSlaveOperator	= &m_opSlave;
		m_pThreadsData[i].dwThreadNo		= i;
		m_pThreadsData[i].dwTotalThreads	= m_dwSlaveThreads;
		m_pThreadsData[i].hThreadFinished	= ::CreateEvent(NULL, FALSE, FALSE, NULL);
		m_pThreadsData[i].hStartEvent		= ::CreateEvent(NULL, FALSE, FALSE, NULL);;
		m_pThreadsData[i].hExitSemaphore	= m_hThreadsExitSemaphore;
		m_pThreadsData[i].llCounterFreq		= m_llCounterFreq;

		m_phThreadsFinished[i]	= m_pThreadsData[i].hThreadFinished;
		m_phThreadsStart[i]		= m_pThreadsData[i].hStartEvent;
	}

	// create threads
	DWORD dwThreadID;

	m_phThreads = new HANDLE[m_dwSlaveThreads];

	for (i = 0; i < m_dwSlaveThreads; ++i) {
		m_phThreads[i] = ::CreateThread(
								NULL, 
								0, 
								CGpgaEngine::SlaveThread, 
								(LPVOID)(m_pThreadsData+i), 
								0, 
								&dwThreadID);

	}

	// let's calculate the init time
	::QueryPerformanceCounter(&m_llEndTime);
	m_llTotalMasterRunTime.QuadPart = m_llEndTime.QuadPart - m_llStartTime.QuadPart;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Destructor

CGpgaEngine::~CGpgaEngine()
{
	DWORD i;

	// stop the threads
	::ReleaseSemaphore(m_hThreadsExitSemaphore, m_dwSlaveThreads, NULL);
	::WaitForMultipleObjects(m_dwSlaveThreads, m_phThreads, TRUE, INFINITE);

	// close hThreadFinished events' handles and thread handles
	for (i = 0; i < m_dwSlaveThreads; ++i) {
		::CloseHandle(m_phThreadsStart[i]);
		::CloseHandle(m_phThreadsFinished[i]);
		::CloseHandle(m_phThreads[i]);
	}

	// delete stuff
	delete [] m_pThreadsData;
	delete [] m_phThreads;
	delete [] m_phThreadsFinished;
	delete [] m_phThreadsStart;

	// close exit semaphore handle
	::CloseHandle(m_hThreadsExitSemaphore);
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- runs the engine for specified number of iterations (max)
//
// DWORD dwIterations	- max numeber of iterations to run (master and/or
//						  slave operators may stop the engine sooner by 
//						  returning 0 from their Run function
//
// returns DWORD	- total number of iterations run

DWORD CGpgaEngine::Run(DWORD dwIterations)
{
	DWORD	i;
	DWORD	j;

	for (i = 0; i < m_dwSlaveThreads; ++i) {
		// reset master stop flag in the threads
		m_pThreadsData[i].bStopMaster = FALSE;
	}

	for (i = 0; i < dwIterations; ++i) {

		// release slave threads, and wait for them to finish
		for (j = 0; j < m_dwSlaveThreads; ++j) {
			::SetEvent(m_phThreadsStart[j]);
		}
		::WaitForMultipleObjects(m_dwSlaveThreads, m_phThreadsFinished, TRUE, INFINITE);

		// check if any slave thread signaled that we should stop
		for (j = 0; j < m_dwSlaveThreads; ++j) {
			if (m_pThreadsData[j].bStopMaster) goto ExitLoop;
		}

		// start the timer...
		::QueryPerformanceCounter(&m_llStartTime);
		// now, apply the master operator
		if (m_opMaster.Run() == 0) {
			::QueryPerformanceCounter(&m_llEndTime);
			m_llTotalMasterRunTime.QuadPart += m_llEndTime.QuadPart - m_llStartTime.QuadPart;
			goto ExitLoop;
		}
		::QueryPerformanceCounter(&m_llEndTime);
		m_llTotalMasterRunTime.QuadPart += m_llEndTime.QuadPart - m_llStartTime.QuadPart;
	}
ExitLoop:
	return (i < m_dwSlaveThreads ? i+1 : i);
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// GetStatistics	- returns the statistics for the previous run
//
// LARGE_INTEGER* pllMasterTime	- receives the master total running time (in msec)
//
// returns DWORD*	- pointer to DWORDs array; i-th DWORD represents the total
//					  running time for i-th thread (in milliseconds)

DWORD* CGpgaEngine::GetStatistics(LARGE_INTEGER* pllMasterTime)
{
	DWORD*	pThreadTimes = new DWORD[m_dwSlaveThreads];

	pllMasterTime->QuadPart = m_llTotalMasterRunTime.QuadPart / m_llCounterFreq.QuadPart;

	for (DWORD i = 0; i < m_dwSlaveThreads; ++i) {
		pThreadTimes[i] = (DWORD)m_pThreadsData[i].llTotalThreadRunTime.QuadPart;

	}

	return pThreadTimes;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// SlaveThread	- slave thread(s) function

DWORD WINAPI CGpgaEngine::SlaveThread(LPVOID lpThreadData)
{
	// get thread data
	THREAD_DATA* pThreadData = (THREAD_DATA*)lpThreadData;
	// get wait handles
	HANDLE		phWaitHandles[2] = { pThreadData->hStartEvent, 
									pThreadData->hExitSemaphore};

	// we need these for statistics
	LARGE_INTEGER	llStartTime;
	LARGE_INTEGER	llEndTime;

	while(::WaitForMultipleObjects(2, phWaitHandles, FALSE, INFINITE) == WAIT_OBJECT_0) {

		::QueryPerformanceCounter(&llStartTime);

		// start event was set, run the slave operator
		if (pThreadData->pSlaveOperator->
			Run(pThreadData->dwThreadNo, pThreadData->dwTotalThreads) == 0) {

			pThreadData->bStopMaster = TRUE;
		}

		// finished with the slave operator, signal to the master
		::QueryPerformanceCounter(&llEndTime);
		pThreadData->llTotalThreadRunTime.QuadPart += (llEndTime.QuadPart - llStartTime.QuadPart) / pThreadData->llCounterFreq.QuadPart;

		::SetEvent(pThreadData->hThreadFinished);
	}

	// exit semaphore was set, exit
	return 1;
}

//////////////////////////////////////////////////////////////////////////////
