//////////////////////////////////////////////////////////////////////////////
// GpgaNQueenOperators.cpp - operator classes for n-Queen problem
//
// This program uses GPGA engine to solve n-Queen problem (place N queens on
// an NxN chessboard in a such way that no queens attack each other)

#include "stdafx.h"
#include "math.h"
#include "GpgaIndividual.h"
#include "GpgaPopulation.h"
#include "GpgaNQueenOperators.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


//////////////////////////////////////////////////////////////////////////////
// Initialization operator class

//////////////////////////////////////////////////////////////////////////////
// constructor

CNQueenInitOperator::CNQueenInitOperator(CPopulation<WORD, float>& Population, COperatorBase& EvalOperator)
:	m_Population(Population), 
	m_EvalOperator(EvalOperator)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// destructor

CNQueenInitOperator::~CNQueenInitOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CNQueenInitOperator::Run()
{
	// init random number generator
	srand(::GetTickCount());

	CreateRandomIndividuals();

	// evaluate the entire popualtion
	m_EvalOperator.Run();

	return 1;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// CreateRandomIndividuals	- creates random chromosomes for all individuals
//							  in the population

void CNQueenInitOperator::CreateRandomIndividuals()
{
	// get population size
	DWORD dwPopulationSize = m_Population.GetPopulationSize();
	// get number of queens
	DWORD dwQueens = (m_Population.GetIndividualPtr(0))->GetChromosomeDim();

	DWORD i;
	DWORD j;

	for (i = 0; i < dwPopulationSize; ++i) {
		// for each individual...
		CIndividual<WORD, float>* pInd = m_Population.GetIndividualPtr(i);

		// ... first create default n-tuple (0, 1, 2, ...) ...
		WORD* pChromosome = pInd->GetChromosomePtr();
		for (j = 0; j < dwQueens; ++j) {
			pChromosome[j] = (WORD)j;
		}

		// ... and then tumble it a little bit
		DWORD dwPos1;
		DWORD dwPos2;
		WORD wTmp;

		for (j = 0; j < dwQueens; ++j) {
			// select two different positions to swap...
			dwPos1 = rand32() % dwQueens;
			while ((dwPos2 = rand32() % dwQueens) == dwPos1);

			// ... and then swap them
			wTmp = pChromosome[dwPos1];
			pChromosome[dwPos1] = pChromosome[dwPos2];
			pChromosome[dwPos2] = wTmp;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Slave operator class

//////////////////////////////////////////////////////////////////////////////

CNQueenSlaveOperator::CNQueenSlaveOperator(COperatorBase& Selection, DWORD dwSlaveIterations)
:	m_Selection(Selection),
	m_dwSlaveIterations(dwSlaveIterations)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////

CNQueenSlaveOperator::~CNQueenSlaveOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator, runs slave operator (for 
//		  single thread config)
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CNQueenSlaveOperator::Run()
{
	return Run(0, 1);
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator, runs slave operator (for 
//		  multi-thread config)
//
// DWORD dwThreadNo		- thread number (zero-based)
// DWORD dwTotalThreads	- total number of threads
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CNQueenSlaveOperator::Run(DWORD dwThreadNo, DWORD dwTotalThreads)
{
	for (DWORD i = 0; i < m_dwSlaveIterations; ++i) {
		if (m_Selection.Run() == 0) return 0;
	}

	return 1;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Evaluation operator class
//
// This operator counts the number of conflicting queens on the board

//////////////////////////////////////////////////////////////////////////////
// constructor

CNQeenEvalOperator::CNQeenEvalOperator(CPopulation<WORD, float>& Population)
:	m_Population(Population)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// destructor

CNQeenEvalOperator::~CNQeenEvalOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator, evaluates the entire population
//
// returns DWORD	- 1 for success, 0 otherwise

DWORD CNQeenEvalOperator::Run()
{
	DWORD	dwRet = 1;
	for (DWORD i = 0; i < m_Population.GetPopulationSize(); ++i) {
		dwRet &= Run(reinterpret_cast<CIndividual<class ChromClass, class FitClass>*>(m_Population.GetIndividualPtr(i)));
	}

	return dwRet;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator, evaluates the given individual
//
// This evaluation function is based on Horowitz's method from "Fundamentals 
// of Computer Algorithms" by Ellis Horowitz and Sartaj Salami, and simplified
// to O(n) as proposed by Kelly D. Crawford in the paper "Solving the n-Queens
// Problem Using Genetic Algorithms"
//
// returns DWORD	- 1 for success, 0xFFFFFFFF otherwise

DWORD CNQeenEvalOperator::Run(CIndividual<class ChromClass, class FitClass>* pIndividual)
{
	CIndividual<WORD, float>* pInd = reinterpret_cast<CIndividual<WORD, float>*>(pIndividual);

	// these are grid diagonals
	DWORD*	pLeftDiagonal = NULL;
	DWORD*	pRightDiagonal = NULL;
	DWORD	dwDiagonals;

	// chromosome pointer
	WORD*	pChromosome;
	DWORD	dwQueens;

	DWORD	i;

	float	fSum;
	DWORD	dwCounter;

	try {
		// get the number of diagonals, and allocate them
		dwQueens = pInd->GetChromosomeDim();
		dwDiagonals = 2 *  dwQueens - 1;

		pLeftDiagonal = new DWORD[dwDiagonals];
		pRightDiagonal = new DWORD[dwDiagonals];
		::ZeroMemory(pLeftDiagonal, dwDiagonals * sizeof(DWORD));
		::ZeroMemory(pRightDiagonal, dwDiagonals * sizeof(DWORD));

		/// get chromosome data pointer
		pChromosome = pInd->GetChromosomePtr();

		// calculate the fitness
		for (i = 0; i < dwQueens; ++i) {
			++pLeftDiagonal[i + pChromosome[i]];
			++pRightDiagonal[dwQueens - i + pChromosome[i] - 1];
		}

		fSum = 0.0;

		for (i = 0; i < dwDiagonals; ++i) {
			dwCounter = 0;

			if (pLeftDiagonal[i] > 1)	dwCounter += pLeftDiagonal[i] - 1;
			if (pRightDiagonal[i] > 1)	dwCounter += pRightDiagonal[i] - 1;

			fSum += (float)dwCounter / (float)(dwQueens - abs(i + 1  - dwQueens));
		}

		pInd->SetFitness(-fSum);

		// cleanup
		delete [] pLeftDiagonal;
		delete [] pRightDiagonal;

		return 1;
	} catch (...) {

		// cleanup
		if (pLeftDiagonal) delete [] pLeftDiagonal;
		if (pRightDiagonal) delete [] pRightDiagonal;

		return 0xFFFFFFFF;
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Selection operator class
//
// Performs a 3-way tournamenet selection. Intended for use in slave threads, 
// therefore, it's thread safe (locks individuals for use)

//////////////////////////////////////////////////////////////////////////////
// Constructor

CNQueenSelectionOperator::CNQueenSelectionOperator(CPopulation<WORD, float>& Population, CEvalOperatorBase& EvalOperator)
:	m_Population(Population), 
	m_EvalOperator(EvalOperator)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Destructor

CNQueenSelectionOperator::~CNQueenSelectionOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator
//
// returns DWORD	- positive number for success, 0xFFFFFFFF otherwise

DWORD CNQueenSelectionOperator::Run()
{
	CIndividual<WORD, float>* pIndividual1;
	CIndividual<WORD, float>* pIndividual2;
	CIndividual<WORD, float>* pIndividual3;
	CIndividual<WORD, float>* pTmpInd;

	DWORD	dwChromosome1;
	DWORD	dwChromosome2;
	DWORD	dwChromosome3;

	try {
		// select first random chromosome index
		do {
			dwChromosome1 = (DWORD) (rand32() % m_Population.GetPopulationSize());
		} while(!m_Population.TryLockIndividual(dwChromosome1));

		// select second random chromosome index 
		// must be different from the first one - we ensure this by 
		// thread-safe access to individuals
		do {
			dwChromosome2 = (DWORD) (rand32() % m_Population.GetPopulationSize());
		} while(!m_Population.TryLockIndividual(dwChromosome2));

		// select third random chromosome index
		// must be different from the previous two - we ensure this by 
		// thread-safe access to individuals
		do {
			dwChromosome3 = (DWORD) (rand32() % m_Population.GetPopulationSize());
		} while(!m_Population.TryLockIndividual(dwChromosome3));

		// now, sort the chromosomes
		pIndividual1 = m_Population.GetIndividualPtr(dwChromosome1);
		pIndividual2 = m_Population.GetIndividualPtr(dwChromosome2);
		pIndividual3 = m_Population.GetIndividualPtr(dwChromosome3);

		if (pIndividual1->GetFitness() < pIndividual2->GetFitness()) {
			pTmpInd = pIndividual1;
			pIndividual1 = pIndividual2;
			pIndividual2 = pTmpInd;
		}

		if (pIndividual2->GetFitness() < pIndividual3->GetFitness()) {
			pTmpInd = pIndividual2;
			pIndividual2 = pIndividual3;
			pIndividual3 = pTmpInd;
		}

		if (pIndividual1->GetFitness() < pIndividual2->GetFitness()) {
			pTmpInd = pIndividual1;
			pIndividual1 = pIndividual2;
			pIndividual2 = pTmpInd;
		}

		// perform the crossover
		CNQueenCrossoverOperator opCrossover(pIndividual1, pIndividual2, pIndividual3);
		if (opCrossover.Run() == 2) {
			// we must evaluate the second individual as well
			m_EvalOperator.Run(reinterpret_cast<CIndividual<class ChromClass, class FitClass>*>(pIndividual2));
		}
		m_EvalOperator.Run(reinterpret_cast<CIndividual<class ChromClass, class FitClass>*>(pIndividual3));

		// unlock the individuals
		m_Population.UnlockIndividual(dwChromosome1);
		m_Population.UnlockIndividual(dwChromosome2);
		m_Population.UnlockIndividual(dwChromosome3);

		if ((pIndividual1->GetFitness() == 0) || 
			(pIndividual2->GetFitness() == 0) || 
			(pIndividual3->GetFitness() == 0)) {
			return 0;
		} else {
			return 1;
		}
	} catch (...) {

		// unlock the individuals
		m_Population.UnlockIndividual(dwChromosome1);
		m_Population.UnlockIndividual(dwChromosome2);
		m_Population.UnlockIndividual(dwChromosome3);
		return 0xFFFFFFFF;
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Crossover operator class
//
// Performs a custom n-Queen crossover

//////////////////////////////////////////////////////////////////////////////
// Constructor

CNQueenCrossoverOperator::CNQueenCrossoverOperator(CIndividual<WORD, float>* pParent1, CIndividual<WORD, float>* pParent2, CIndividual<WORD, float>* pOffspring)
:	m_pParent1(pParent1), 
	m_pParent2(pParent2), 
	m_pOffspring(pOffspring)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Destructor

CNQueenCrossoverOperator::~CNQueenCrossoverOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator
//
// returns DWORD	- 1 for success, 0xFFFFFFFF otherwise

DWORD CNQueenCrossoverOperator::Run()
{
	try {
		DWORD	dwQueens	= m_pParent1->GetChromosomeDim();
		WORD*	pwChrom1	= m_pParent1->GetChromosomePtr();
		WORD*	pwChrom2	= m_pParent2->GetChromosomePtr();
		WORD*	pwChrom3	= m_pOffspring->GetChromosomePtr();

		WORD	wTmp;

		DWORD	dwPos1;
		DWORD	dwPos2;
		DWORD	dwTmpPos;

		// crossover v3 :-)
		DWORD i;
		DWORD j;
		DWORD dwResult = 1;

		// ... first create default n-tuple (0, 1, 2, ...) ...
		for (i = 0; i < dwQueens; ++i) {
			pwChrom3[i] = (WORD)i;
		}

		// ... and then tumble it a little bit
		for (i = 0; i < dwQueens; ++i) {
			// select two different positions to swap...
			dwPos1 = rand32() % dwQueens;
			while ((dwPos2 = rand32() % dwQueens) == dwPos1);

			// ... and then swap them
			wTmp = pwChrom3[dwPos1];
			pwChrom3[dwPos1] = pwChrom3[dwPos2];
			pwChrom3[dwPos2] = wTmp;
		}

		// parents are equal?
		if (memcmp(pwChrom1, pwChrom2, dwQueens * sizeof(WORD)) == 0) {
			// yes, create a random chromosome for the second individual
			// ... first create default n-tuple (0, 1, 2, ...) ...
			for (i = 0; i < dwQueens; ++i) {
				pwChrom2[i] = (WORD)i;
			}

			// ... and then tumble it a little bit
			for (i = 0; i < dwQueens; ++i) {
				// select two different positions to swap...
				dwPos1 = rand32() % dwQueens;
				while ((dwPos2 = rand32() % dwQueens) == dwPos1);

				// ... and then swap them
				wTmp = pwChrom2[dwPos1];
				pwChrom2[dwPos1] = pwChrom2[dwPos2];
				pwChrom2[dwPos2] = wTmp;
			}
			dwResult = 2;
		}

		// perform the crossover
		// see which positions in parents coincide
		for (i = 0; i < dwQueens; ++i) {
			if (pwChrom1[i] == pwChrom2[i]) {
				for (j = 0; j < dwQueens; ++j) {
					if (pwChrom3[j] == pwChrom1[i]) break;
				}

				wTmp = pwChrom3[i];
				pwChrom3[i] = pwChrom3[j];
				pwChrom3[j] = wTmp;
			}
		}
		return dwResult;

//////////////////////////////////////////////////////////////////////////////


//		// select two different points in the tuple...
//		dwPos1 = rand32() % dwQueens;
//		while ((dwPos2 = rand32() % dwQueens) == dwPos1);
//
//		// swap positions in the tuple
//		wTmp			= pwChrom2[dwPos1];
//		pwChrom2[dwPos1]= pwChrom2[dwPos2];
//		pwChrom2[dwPos2]= wTmp;
//
//		// select two different points in the tuple...
//		dwPos1 = rand32() % dwQueens;
//		while ((dwPos2 = rand32() % dwQueens) == dwPos1);
//
//		// swap positions in the tuple
//		wTmp			= pwChrom3[dwPos1];
//		pwChrom3[dwPos1]= pwChrom3[dwPos2];
//		pwChrom3[dwPos2]= wTmp;


//////////////////////////////////////////////////////////////////////////////


//		// we'll tumble 2nd and 3rd individual...
//
//		// 2nd...
//		// select two random points within the tuple
//		dwPos1 = rand32() % dwQueens;
//		while ((dwPos2 = rand32() % dwQueens) == dwPos1);
//
//		// swap points if needed
//		if (dwPos1 > dwPos2) {
//			dwTmpPos = dwPos1;
//			dwPos1 = dwPos2;
//			dwPos2 = dwTmpPos;
//		}
//
//		// and now invert the part of the tuple
//		while (dwPos1 < dwPos2) {
//			wTmp				= pwChrom2[dwPos1];
//			pwChrom2[dwPos1]	= pwChrom2[dwPos2];
//			pwChrom2[dwPos2]	= wTmp;
//
//			++dwPos1;
//			--dwPos2;
//		}
//
//		// 3rd...
//		// select two random points within the tuple
//		dwPos1 = rand32() % dwQueens;
//		while ((dwPos2 = rand32() % dwQueens) == dwPos1);
//
//		// swap points if needed
//		if (dwPos1 > dwPos2) {
//			dwTmpPos = dwPos1;
//			dwPos1 = dwPos2;
//			dwPos2 = dwTmpPos;
//		}
//
//		// and now invert the part of the tuple
//		while (dwPos1 < dwPos2) {
//			wTmp				= pwChrom3[dwPos1];
//			pwChrom3[dwPos1]	= pwChrom3[dwPos2];
//			pwChrom3[dwPos2]	= wTmp;
//
//			++dwPos1;
//			--dwPos2;
//		}
//
//		// done!
//		return 1;
	} catch (...) {
		return 0xFFFFFFFF;
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Mutation operator class
//
// Performs a simple n-Queen mutation on a population

//////////////////////////////////////////////////////////////////////////////
// Constructor

CNQueenMutationOperator::CNQueenMutationOperator(CPopulation<WORD, float>& Population, CEvalOperatorBase& EvalOperator, double dChromosomeMutationRate)
:	m_Population(Population), 
	m_EvalOperator(EvalOperator), 
	m_dChromosomeMutationRate(dChromosomeMutationRate)
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Destructor

CNQueenMutationOperator::~CNQueenMutationOperator()
{
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Run	- entry point for the operator
//
// returns DWORD	- 1 for success, 0xFFFFFFFF otherwise

DWORD CNQueenMutationOperator::Run()
{
	try {
		// determine the number of mutating chromosomes
		// we need this +0.5 in order to ensure correct truncation 
		// of the float value
		DWORD dwMutatingChromosomes;
		dwMutatingChromosomes = (DWORD)(m_Population.GetPopulationSize() * 
										m_dChromosomeMutationRate + 0.5);

		DWORD	dwBestIndividual = m_Population.GetBestIndividualIndex();
 
		// now, select dwMutatingChromosomes chromosomes to mutate
		DWORD	dwRandChromosome;
		DWORD	dwPos1;
		DWORD	dwPos2;

		CIndividual<WORD, float>*	pMutatingInd;
		DWORD						dwQueens;
		WORD*						pMutatingIndChrom;
		WORD						wTmp;

		for (DWORD i = 0; i < dwMutatingChromosomes; ++i) {
			// select random chromosome 
			// (we must not select the best individual)
			do {
				dwRandChromosome = rand32() % m_Population.GetPopulationSize();
			} while (dwRandChromosome == dwBestIndividual);

			pMutatingInd		= m_Population.GetIndividualPtr(dwRandChromosome);
			dwQueens			= pMutatingInd->GetChromosomeDim();

			// select two different points in the tuple...
			dwPos1 = rand32() % dwQueens;
			while ((dwPos2 = rand32() % dwQueens) == dwPos1);

			// ... mutate the individual...
			pMutatingIndChrom	= pMutatingInd->GetChromosomePtr();

			// swap positions in the tuple
			wTmp						= pMutatingIndChrom[dwPos1];
			pMutatingIndChrom[dwPos1]	= pMutatingIndChrom[dwPos2];
			pMutatingIndChrom[dwPos2]	= wTmp;

			// ... and evaluate the new individual
			m_EvalOperator.Run(reinterpret_cast<CIndividual<class ChromClass, class FitClass>*>(pMutatingInd));
			if (pMutatingInd->GetFitness() == 0) {
				return 0;
			}
		}

		// done! return.
		return 1;

	} catch (...) {
		return 0xFFFFFFFF;
	}
}

//////////////////////////////////////////////////////////////////////////////

