//////////////////////////////////////////////////////////////////////////////
// GpgaPopulation.h	- population template class
//////////////////////////////////////////////////////////////////////////////

// max size of the population
#define MAX_POP		32767

//////////////////////////////////////////////////////////////////////////////
// CPopulation - population template class
//
// This class is used to handle a GA population. Class T must be a class 
// of type CIndividual, or derived from it.


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// class declaration

template <class ChromClass, class FitClass>
class CPopulation {
	public:
		// constructors
		CPopulation(DWORD dwPopulationSize, DWORD dwChromosomeDimension);
		CPopulation(CPopulation<ChromClass, FitClass>& Population, DWORD dwPopChunkStartIndex, DWORD dwPopChunkSize);

		// destructor
		~CPopulation();

		//////////////////////////////////////////////////////////////////////
		// operations

		// returns pointer to the individual at the given index (zero-based)
		CIndividual<ChromClass, FitClass>* GetIndividualPtr(DWORD dwIndex);

		// returns population size
		DWORD GetPopulationSize();

		// returns the best individual index
		DWORD GetBestIndividualIndex();

		// returns the pointer to the best individual
		CIndividual<ChromClass, FitClass>* GetBestIndividualPtr();

		// thread safe stuff
		BOOL TryLockIndividual(DWORD dwIndex);
		void UnlockIndividual(DWORD dwIndex);

	private:
		CIndividual<ChromClass, FitClass>*	m_pPopulation;
		// size of the population
		DWORD	m_dwPopulationSize;

		BOOL	m_bCleanup;

		// stuff needed for thread-safe access to the population individuals

		// 'individuals-in-use' flag array
		BOOL*	m_pbIndividualFlags;

		// critical section for accessing m_pbIndividualFlags array
		CRITICAL_SECTION	m_csFlags;
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// class definition (this *should* go to a .cpp file, but @#@!$ VC has a bug
// (see MSDN article Q239436)


//////////////////////////////////////////////////////////////////////////////
// constructor	- creates the population object and initializes it

template <class ChromClass, class FitClass>
CPopulation<ChromClass, FitClass>::CPopulation(DWORD dwPopulationSize, DWORD dwChromosomeDimension)
:	m_bCleanup(1)
{
	m_dwPopulationSize = dwPopulationSize;
	// we can't have population size greater than MAX_POP
	if (m_dwPopulationSize > MAX_POP) m_dwPopulationSize = MAX_POP;

	m_pPopulation = new CIndividual<ChromClass, FitClass>[m_dwPopulationSize];
	for (DWORD i = 0; i < m_dwPopulationSize; ++i) {
		m_pPopulation[i].AllocChromosome(dwChromosomeDimension);
	}

	// init thread-safe stuff
	::InitializeCriticalSection(&m_csFlags);
	m_pbIndividualFlags = new BOOL[m_dwPopulationSize];
	::ZeroMemory(m_pbIndividualFlags, m_dwPopulationSize * sizeof(BOOL));
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// constructor	- creates the population object
// 
// Note: This constructor doesn't allocate individuals. Use this constructor
//       when you need to get a part of the population

template <class ChromClass, class FitClass>
CPopulation<ChromClass, FitClass>::CPopulation(CPopulation<ChromClass, FitClass>& Population, DWORD dwPopChunkStartIndex, DWORD dwPopChunkSize)
:	m_bCleanup(0)
{
	if (dwPopChunkStartIndex+dwPopChunkSize > Population.GetPopulationSize()) {
		m_dwPopulationSize = Population.GetPopulationSize() - dwPopChunkStartIndex;
	} else {
		m_dwPopulationSize = dwPopChunkSize;
	}

	m_pPopulation = Population.GetIndividualPtr(dwPopChunkStartIndex);

	// init thread-safe stuff
	m_csFlags = Population.m_csFlags;
	m_pbIndividualFlags = (Population.m_pbIndividualFlags)+dwPopChunkStartIndex;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// destructor	- cleanup:
//					- deletes the population object

template <class ChromClass, class FitClass>
CPopulation<ChromClass, FitClass>::~CPopulation()
{
	if (m_bCleanup) {
		delete [] m_pPopulation;
		delete [] m_pbIndividualFlags;
		::DeleteCriticalSection(&m_csFlags);
	}
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// GetIndividualPtr	- returns pointer to the individual at the given index 
//					  (zero-based); provides safe access to the population 
//					  members
//
// DWORD dwIndex	- index of the individual to be retreived
//
// returns T*		- pointer to the individual

template <class ChromClass, class FitClass>
CIndividual<ChromClass, FitClass>* CPopulation<ChromClass, FitClass>::GetIndividualPtr(DWORD dwIndex)
{
	if (dwIndex >= m_dwPopulationSize) return NULL;

	return (CIndividual<ChromClass, FitClass>*)(m_pPopulation+dwIndex);
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// GetPopulationSize	- returns population size
//
// returns DWORD		- population size

template <class ChromClass, class FitClass>
DWORD CPopulation<ChromClass, FitClass>::GetPopulationSize()
{
	return m_dwPopulationSize;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// GetBestIndividual	- returns the pointer to the best individual
//
// returns ChromClass*	- pointer to the best individual

template <class ChromClass, class FitClass>
CIndividual<ChromClass, FitClass>* CPopulation<ChromClass, FitClass>::GetBestIndividualPtr()
{
	// m_dwBestIndividualIndex is not set, search the population
	CIndividual<ChromClass, FitClass>* pBestIndividual;

	pBestIndividual = &(m_pPopulation[0]);

	for (DWORD i = 1; i < m_dwPopulationSize; ++i) {
		if (m_pPopulation[i].GetFitness() > pBestIndividual->GetFitness()) {
			pBestIndividual = &(m_pPopulation[i]);
		}
	}

	return pBestIndividual;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// GetBestIndividualIndex	- returns the best individual index

template <class ChromClass, class FitClass>
DWORD CPopulation<ChromClass, FitClass>::GetBestIndividualIndex()
{
	DWORD dwBestIndividualIndex = 0;

	for (DWORD i = 1; i < m_dwPopulationSize; ++i) {
		if (m_pPopulation[i].GetFitness() > 
			m_pPopulation[dwBestIndividualIndex].GetFitness()) {

			dwBestIndividualIndex = i;
		}
	}

	return dwBestIndividualIndex;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// TryLockIndividual	- attempts to lock the individual at given index
//
// DWORD dwIndex	- index of the individual to lock
//
// returns BOOL	- TRUE for successful lock, FALSE otherwise

template <class ChromClass, class FitClass>
BOOL CPopulation<ChromClass, FitClass>::TryLockIndividual(DWORD dwIndex)
{
	BOOL bRetVal = FALSE;

	if (dwIndex >= m_dwPopulationSize) return FALSE;

	::EnterCriticalSection(&m_csFlags);
	if (!m_pbIndividualFlags[dwIndex]) {
		m_pbIndividualFlags[dwIndex] = TRUE;
		bRetVal = TRUE;
	}
	::LeaveCriticalSection(&m_csFlags);

	return bRetVal;
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// UnlockIndividual	- unlocks the individual at given index
//
// DWORD dwIndex	- index of the individual to unlock
//

template <class ChromClass, class FitClass>
void CPopulation<ChromClass, FitClass>::UnlockIndividual(DWORD dwIndex)
{
	if (dwIndex >= m_dwPopulationSize) return;

	::EnterCriticalSection(&m_csFlags);
	m_pbIndividualFlags[dwIndex] = FALSE;
	::LeaveCriticalSection(&m_csFlags);
}

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
