//////////////////////////////////////////////////////////////////////////////
// GpgaOperatorBase.h	- base class for operator sets
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Classes derived from this one are used to encapsulate operators 
// (not just GA operators!) that will be applied by the GPGA engine.
//
// GPGA engine applies sets in the following order:
//    1. Initialization operator is run. This operator performs steps such as
//       random population creation, initial evaluation, etc.
//
//    2. Slave operator set is run in several threads. Slave operator set
//       includes at least evaluation operator and may include other GA 
//		 operators.
//
//    3. When slaves have finished, a master operator set is run.
//
// Steps 2. and 3. are repeated until a stopping criterion is met (for this 
// implementation, a specified number of master iterations)
//
// Note: Run function from slave and master classes may return 0 to signal to 
//       GPGA engine that it should stop with iterations (e.g. when a solution
//       has been found, time limit hit, etc.)

class COperatorBase {
	public:
		COperatorBase(){};
		virtual ~COperatorBase(){};

		virtual DWORD Run() = 0;
};

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// CEvalOperatorBase	- base class for evaluation operator classes

class CEvalOperatorBase : public COperatorBase {
	public:
		CEvalOperatorBase(){};
		virtual ~CEvalOperatorBase(){};

		virtual DWORD Run() = 0;
		virtual DWORD Run(CIndividual<class ChromClass, class FitClass>* pIndividual) = 0;
};

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// CSlaveOperatorBase	- base class for slave operator classes

class CSlaveOperatorBase : public COperatorBase {
	public:
		CSlaveOperatorBase(){};
		virtual ~CSlaveOperatorBase(){};

		virtual DWORD Run() = 0;
		virtual DWORD Run(DWORD dwThreadNo, DWORD dwTotalThreads) = 0;
};

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// Exception class used internally by the operator classes to indicate errors

class OperatorException {};


