//////////////////////////////////////////////////////////////////////////////
// GpgaRealFn.cpp : Defines the entry point for the console application.
//
// This program uses GPGA engine, and default operators to optimize
// real functions (see CRealFnEvalOperator class for functions details)

#include "stdafx.h"
#include "GpgaIndividual.h"
#include "GpgaPopulation.h"
#include "GpgaRealFnOperators.h"
#include "GpgaEngine.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//////////////////////////////////////////////////////////////////////////////
#define IS_SPACE(c)	((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\r')
//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////
// keywords

#define KEYWORDS	9
CHAR* Keywords[] = {
	"function_id", 
	"population_size", 
	"slave_iterations", 
	"slave_threads", 
	"master_iterations", 
	"lower_limits", 
	"upper_limits", 
	"mutation_rate", 
	"optimize_flag"
};

//////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////

int main(int argc, char* argv[])
{
	// turn on memory allocation tracking
#	ifdef _DEBUG
		int nDbgFlag = ::_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG);
		nDbgFlag |= _CRTDBG_LEAK_CHECK_DF;
		::_CrtSetDbgFlag(nDbgFlag);
#	endif

	// we need two cmd-line parameters, input and output files
	if (argc != 3) {
		printf("Usage: GpgaRealFn.exe <input file> <output file>\n For details see sample input file.\n");
		return -1;
	}

	//////////////////////////////////////////////////////////////////////////
	// GA variables

	int		nFunctionID;
	DWORD	dwFunctionDim;
	DWORD	dwPopulationSize;
	DWORD	dwSlaveIterations;
	DWORD	dwSlaveThreads;
	DWORD	dwMasterIterations;
	double	dBitMutationRate;
	int		nOptimizeFlag;

	double*	pdLowerLimits;
	double*	pdUpperLimits;

	//////////////////////////////////////////////////////////////////////////


	//////////////////////////////////////////////////////////////////////////
	// Read config file and set variables

	// read the input file
	HANDLE	hInFile;
	CHAR*	pszFileData;
	DWORD	dwFileSize;

	if ((hInFile = ::CreateFile(
					argv[1], 
					GENERIC_READ, 
					FILE_SHARE_READ, 
					NULL, 
					OPEN_EXISTING, 
					FILE_FLAG_SEQUENTIAL_SCAN, 
					NULL)) == NULL) {

		printf("Unable to open input file!\n");
		return -1;
	}

	dwFileSize = ::GetFileSize(hInFile, NULL);
	pszFileData = new CHAR[dwFileSize + 1];
	::ZeroMemory(pszFileData, dwFileSize + 1);

	::ReadFile(hInFile, pszFileData, dwFileSize, &dwFileSize, NULL);

	// let's parse the data now
	CHAR*	pTmp = pszFileData;
	CHAR	szToken[256];
	BOOL	bSkipToEOL = FALSE;

	DWORD	i;
	DWORD	j;

	while (pTmp[0] != '\0') {

		if (bSkipToEOL) {
			// skip the comments
			if (pTmp[0] == '\n') bSkipToEOL = FALSE;
			++pTmp;
		} else {
			if (pTmp[0] == '#') {
				// comment
				bSkipToEOL = TRUE;
				++pTmp;
			} else if (IS_SPACE(pTmp[0])) {
				++pTmp;
			} else {
				// we have some text...
				sscanf(pTmp, "%s", szToken);

				// ... is it a keyword?
				for (i = 0; i < KEYWORDS; ++i) {
					if (!_strcmpi(szToken, Keywords[i])) break;
				}
				pTmp += strlen(szToken);

				// let's see
				switch (i) {
					// function_id
					case 0	:
						sscanf(pTmp, "%i", &nFunctionID);
						dwFunctionDim = abs(nFunctionID / 1000);
						bSkipToEOL = TRUE;
						break;

					// population_size
					case 1	:
						sscanf(pTmp, "%i", &dwPopulationSize);
						bSkipToEOL = TRUE;
						break;

					// slave_iterations
					case 2	:
						sscanf(pTmp, "%i", &dwSlaveIterations);
						bSkipToEOL = TRUE;
						break;

					// slave_threads
					case 3	:
						sscanf(pTmp, "%i", &dwSlaveThreads);
						bSkipToEOL = TRUE;
						break;

					// master_iterations
					case 4	:
						sscanf(pTmp, "%i", &dwMasterIterations);
						bSkipToEOL = TRUE;
						break;

					// lower_limits
					case 5	:
						pdLowerLimits = new double[dwFunctionDim];
						for (j = 0; j < dwFunctionDim; ++j) {
							while (IS_SPACE(pTmp[0])) ++pTmp;
							sscanf(pTmp, "%lf", &pdLowerLimits[j]);
							while (!IS_SPACE(pTmp[0])) ++pTmp;
						}
						break;

					// upper_limits
					case 6	:
						pdUpperLimits = new double[dwFunctionDim];
						for (j = 0; j < dwFunctionDim; ++j) {
							while (IS_SPACE(pTmp[0])) ++pTmp;
							sscanf(pTmp, "%lf", &pdUpperLimits[j]);
							while (!IS_SPACE(pTmp[0])) ++pTmp;
						}
						break;

					// mutation_rate
					case 7	:
						sscanf(pTmp, "%lf", &dBitMutationRate);
						bSkipToEOL = TRUE;
						break;

					// optimize_flag
					case 8	:
						sscanf(pTmp, "%i", &nOptimizeFlag);
						nOptimizeFlag <= 0 ? nOptimizeFlag = -1 : nOptimizeFlag = 1;
						bSkipToEOL = TRUE;
						break;

					default : bSkipToEOL = TRUE;
				};
			}
		}
	}


	// cleanup file stuff
	delete pszFileData;
	::CloseHandle(hInFile);

	//////////////////////////////////////////////////////////////////////////


	//////////////////////////////////////////////////////////////////////////
	// Create population, operators and engine

	CPopulation<DWORD, double>	Population(dwPopulationSize, dwFunctionDim);

	CRealFnEvalOperator			EvalOperator(
									Population, 
									pdLowerLimits, 
									pdUpperLimits, 
									nFunctionID, 
									nOptimizeFlag);

	CRealFnInitOperator			InitOperator(
									Population, 
									EvalOperator);

	CRealFnSelectionOperator	SelectionOperator(
									Population, 
									EvalOperator);

	CRealFnSlaveOperator		SlaveOperator(
									SelectionOperator, 
									dwSlaveIterations);

	CRealFnMutationOperator		MutationOperator(
									Population, 
									EvalOperator, 
									dBitMutationRate);

	CGpgaEngine					GpgaEngine(
									InitOperator, 
									MutationOperator, 
									SlaveOperator, 
									dwSlaveThreads);

	//////////////////////////////////////////////////////////////////////////


	//////////////////////////////////////////////////////////////////////////
	// Run the engine, and create output file with results

	GpgaEngine.Run(dwMasterIterations);

	FILE* hOutputFile;

	CIndividual<DWORD, double>* pBestIndividual;
	double*						pBestIndividualDecoded;
	DWORD*						pThreadsTimes;
	LARGE_INTEGER				llMasterTime;

	pBestIndividual			= Population.GetBestIndividualPtr();
	pBestIndividualDecoded	= EvalOperator.Decode(pBestIndividual);
	pThreadsTimes			= GpgaEngine.GetStatistics(&llMasterTime);

	hOutputFile = fopen(argv[2], "wt");

	fprintf(hOutputFile, "##############################################################################\n");
	fprintf(hOutputFile, "# Function ID:              %i\n", nFunctionID);
	fprintf(hOutputFile, "##############################################################################\n");
	fprintf(hOutputFile, "# Population size:          %i\n", dwPopulationSize);
	fprintf(hOutputFile, "# Mutation probability:     %.4lf\n", dBitMutationRate);
	fprintf(hOutputFile, "# Number of slave iter.:    %i\n", dwSlaveIterations);
	fprintf(hOutputFile, "# Number of slave threads:  %i\n", dwSlaveThreads);
	fprintf(hOutputFile, "# Number of master iter.:   %i\n", dwMasterIterations);
	fprintf(hOutputFile, "# Finding :                 ");
	if (nOptimizeFlag == -1) {
		fprintf(hOutputFile, "minimum value\n");
	} else {
		fprintf(hOutputFile, "maximum value\n");
	}
	fprintf(hOutputFile, "##############################################################################\n");

	fprintf(hOutputFile, "\n\n");
	
	fprintf(hOutputFile, "##############################################################################\n");
	fprintf(hOutputFile, "# Master time: %I64i msec\n", llMasterTime.QuadPart);
	fprintf(hOutputFile, "##############################################################################\n");
	fprintf(hOutputFile, "# Threads' times\n");
	fprintf(hOutputFile, "#\n");

	for (i = 0; i < dwSlaveThreads; ++i) {
		fprintf(hOutputFile, "# Thread %i\n", i);
		fprintf(hOutputFile, "# Total running time: %i msec\n", pThreadsTimes[i]);
		fprintf(hOutputFile, "# Avg. running time: %i msec/master iter.\n", pThreadsTimes[i] / dwMasterIterations);
		fprintf(hOutputFile, (i == dwSlaveThreads - 1 ? "" : "#\n"));
	}
	fprintf(hOutputFile, "##############################################################################\n");

	fprintf(hOutputFile, "\n\n");

	fprintf(hOutputFile, "##############################################################################\n");
	fprintf(hOutputFile, "# Best individual:\n");
	fprintf(hOutputFile, "#\n");
	fprintf(hOutputFile, "# Fitness value: %lf\n", pBestIndividual->GetFitness());
	fprintf(hOutputFile, "# Function value: %lf\n", pBestIndividual->GetFitness() * nOptimizeFlag);
	fprintf(hOutputFile, "#\n");
	fprintf(hOutputFile, "# Function input vector:\n");

	for (i = 0; i < dwFunctionDim; ++i) {
		fprintf(hOutputFile, "# %lf\n", pBestIndividualDecoded[i]);
	}

	fprintf(hOutputFile, "##############################################################################\n");

	fclose(hOutputFile);
	delete pBestIndividualDecoded;
	delete pThreadsTimes;

	//////////////////////////////////////////////////////////////////////////

	// cleanup
	delete [] pdLowerLimits;
	delete [] pdUpperLimits;

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
